#!/usr/bin/env node
'use strict';

// This file is created for our internal testing purposes.

var Parser = require('posix-getopt').BasicParser;
var VERSION = require('../package.json').version;
var basename = require('path').basename;
var controller = require('../lib/controller');
var debug = require('../lib/debug')('wlpn-diagnostics:bin');
var fs = require('fs');
var mkdirp = require('mkdirp');
var path = require('path');
var prettyjson = require('prettyjson');

var optind = 1;
var $0 = basename(process.argv[optind++]);
var command = process.argv[optind++];

var commands = {
  help: help,
  version: version,
  checkAll: controller.checkAll,
};

if (!command) {
  printHelp($0, console.error);
  return exit(1);
}

var commandFn = commands[command];

if (!commandFn) {
  console.error('Unknown action: %s', command);
  printHelp($0, console.error);
  return exit(1);
}

var serverCommandParser = new Parser([
  ':',
  'u:(adminUser)',
  'p:(adminPass)',
  'H:(controllerHost)',
  'P:(controllerPort)',
  'w:(memberHost)',
  'x:(memberPort)',
  'y:(memberSSHUser)',
  'z:(memberSSHPass)',
].join(''), process.argv, optind);

var option;
var options = {};

while ((option = serverCommandParser.getopt()) !== undefined) {
  switch (option.option) {
    case 'u': options.adminUser = option.optarg; break;
    case 'p': options.adminPass = option.optarg; break;
    case 'H': options.controllerHost = option.optarg; break;
    case 'P': options.controllerPort = option.optarg; break;
    case 'w': options.memberHost = option.optarg; break;
    case 'x': options.memberPort = option.optarg; break;
    case 'y': options.memberSSHUser = option.optarg; break;
    case 'z': options.memberSSHPass = option.optarg; break;
    default:
      console.error('Invalid usage at ' + option.optopt);
      return exit(1);
  }
}

commandFn(options, function(err, results) {
  var outputDir = path.join(process.cwd(), 'wlpn-diagnostics-results');
  mkdirp.sync(outputDir);

  debug('Something went incredibly wrong: %s', err);
  logFile({
    logPath: path.join(outputDir, 'controller.results.json.log'),
  }, JSON.stringify(results));
  var prettyLog = prettyjson.render(results);
  console.log(prettyLog);
  logFile({
    logPath: path.join(outputDir, 'controller.results.log'),
  }, prettyLog);
});

function version() {
  console.log('%s', VERSION);
  exit(0);
}

function help(argv, optind, options, cb) {
  try {
    printHelp($0, console.error);
  } catch (err) {
    console.error(err);
    printHelp($0, console.error);
    exit(1);
  }
}

function exit(code) {
  return setImmediate(process.exit, code);
}

function printHelp($0, prn) {
  var USAGE = fs.readFileSync(require.resolve('./wlpn-diagnostics.txt'),
    'utf-8')
    .replace(/%MAIN%/g, $0)
    .trim();
  prn(USAGE);
}

function logFile(options, text) {
  var logPath = options.logPath || path.resolve(process.cwd(),
    'wlpn-diagnostics-output.log');
  fs.writeFile(logPath, text, function(err) {
    if (err) {
      return console.error(err);
    }
  });
};
