// Copyright IBM Corp. 2016. All Rights Reserved.
// Node module: apiconnect-collective-controller-api
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.

'use strict';

var _ = require('lodash');
var Client = require('ssh2').Client;
var async = require('async');
var chalk = require('chalk');
var debug = require('./debug')('wlpn-diagnostics:controller');
var exec = require('child_process').exec;
var fs = require('fs');
var jmx = require('apiconnect-collective-jmx');
var mandatory = require('./util').mandatory;
var nocb = require('./util').nocb;
var os = require('os');
var path = require('path');
var userhome = require('userhome');
var xml2js = require('xml2js');
var CTRL_PATH = userhome('.liberty', 'wlp', 'usr', 'servers', 'controller');
var SHARED_PATH = userhome('.liberty', 'wlp', 'usr', 'shared');

exports.checkAll = checkAll;

var parser = new xml2js.Parser();

function checkAll(options, cb) {
  async.series([
    checkNodeEnv,
    checkLibertyFolder.bind(null, options),
    checkStackGroup.bind(null, options),
    checkNetwork.bind(null, options),
  ], checkAllResults);

  function checkAllResults(err, results) {
    if (err) {
      debug(err);
      return cb(err);
    }
    cb = cb || nocb;
    results = _.reduce(results, function(list, obj) {
      return _.merge(list, obj);
    });
    return cb(null, results);
  };
};

function checkNodeEnv(cb) {
  exec('npm ls -g --depth=0 --json', function(err, stdout, stderr) {
    var npmDeps;
    if (err || stderr){
      npmDeps = chalk.yellow('WARNING: could not get dependencies');
    }

    try {
      npmDeps = JSON.parse(stdout);
    } catch (e) {
      npmDeps = chalk.yellow('WARNING: could not parse npm dependencies');
    }
    return cb(null, {
      nodePath: process.execPath,
      npmDeps: npmDeps,
    });
  });
}

function checkLibertyFolder(options, cb) {
  fs.lstat(CTRL_PATH, function(err, stat) {
    if (err || !stat.isDirectory()) {
      return cb(null, {
        libertyFolder: {
          exists: false,
          msg: chalk.red('ERROR: controller does not exist, please re-setup!'),
          err: err || 'not a directory',
        },
      });
    }

    options.libertyExists = true;

    var tasks = [
      checkServerXml.bind(null, options),
      checkDeployRule.bind(null, options),
    ];

    async.parallel(async.reflectAll(tasks), function(err, results) {
      if (err) {
        debug('something really bad happened: %s', err);
        return cb(err);
      }
      var data = {
        serverXml: results[0],
        deployRuleXml: results[1],
      };

      // merge all results with below return obj
      results = _.merge({}, {
        libertyFolder: {
          exists: true,
        },
      }, data);
      return cb(null, results);
    });
  });
};

function checkServerXml(options, cb) {
  var xmlPath = path.join(CTRL_PATH, 'server.xml');
  fs.readFile(xmlPath, function(err, data) {
    if (err) {
      debug('could not read %s: %s', xmlPath, err);
      return cb(err);
    }
    parser.parseString(data, function(err, xml) {
      if (err) {
        return cb(new Error('server.xml could not be read'));
      }
      var deployVars = {};
      deployVars.deployVarFormat = _.map(xml.server.deployVariable,
        function(ele) {
          var message = 'Looks good';
          var name = ele.$.name;
          if (!/wlpn\./gi.test(name)) {
            message = chalk.red('ERROR: does not have `wlpn.` prefix. Are ' +
              'you a WAS user?');
          }
          return {
            deployVar: name,
            raw: ele.$,
            message: message,
          };
        });
      return cb(null, deployVars);
    });
  });
};

function checkDeployRule(options, cb) {
  var xmlPath = path.join(CTRL_PATH, 'deployRule.xml');
  fs.readFile(xmlPath, function(err, data) {
    if (err) {
      debug('could not read %s: %s', xmlPath, err);
      return cb(err);
    }
    parser.parseString(data, function(err, xml) {
      if (err) {
        return cb(new Error('deployRule.xml could not be read'));
      }
      var deployRuleEle = xml.server.deployRule[0];
      var results = {};

      // Check deployRule parent element
      var ele = deployRuleEle.$;
      if (!/APIC Server Rule/gi.test(ele.id)) {
        results.deployRuleId = {
          value: ele.id,
          message: chalk.yellow('WARNING: deployRule ID does not match ' +
            '`APIC Server Rule`; controller could be old (<v5020)'),
        };
      }
      if (!/Node\.js/gi.test(ele.type)) {
        results.deployRuleType = {
          value: ele.type,
          message: chalk.red('ERROR: deployRule type does not match ' +
            '`Node.js`'),
        };
      }

      // Check <deploy> commands
      try {
        var deployCmd = deployRuleEle.deploy[0];
      } catch (e) {
        // Return error and ontinue on to other checks
        results.deployCmd = chalk.red('ERROR: could not read <deploy> ' +
          'commands');
      }
      if (/(mkdir|wget)/gi.test(deployCmd)) {
        results.wget = chalk.yellow('WARNING: mkdir/wget detected; ' +
          'controller could be old (<v5010)');
      }
      if (/(wlpn-)/gi.test(deployCmd)) {
        results.wlpnFormat = chalk.yellow('WARNING: wlpn- detected; ' +
          'controller could be old (<v5020)');
      }

      // Check invputVariables
      try {
        var inputVar = deployRuleEle.inputVariable;

        inputVar = _.map(inputVar, function(ele) {
          var message = 'Looks good';
          var name = ele.$.name;
          if (/[a-zA-Z]+\.[a-zA-Z]+/gi.test(name)) {
            message = chalk.yellow('WARNING: found dot notation in input ' +
              'variables; controller could be old (<v5020)');
          }
          return {
            inputVar: name,
            message: message,
          };
        });
        results.inputVarFormat = inputVar;
      } catch (e) {
        results.inputVarFormat = chalk.red('ERROR: could not read ' +
          '<inputVariable> elements');
      }

      if (_.isEmpty(results)) {
        results.deployRuleXml = 'deployRule.xml looks good';
      }
      return cb(null, results);
    });
  });
};

function checkStackGroup(options, cb) {
  var APIC_STACKGROUP_PATH = path.join(SHARED_PATH, 'stackGroups', 'apic',
    'packages');
  fs.lstat(APIC_STACKGROUP_PATH, function(err, stat) {
    if (err || !stat.isDirectory()) {
      return cb(null, {
        apicStackGroupExists: {
          exists: false,
          msg: chalk.red('ERROR: deploy command might have failed or you ' +
            'didn\'t run it'),
          err: err || 'not a directory',
        },
      });
    }
    return cb(null, {
      apicStackGroupExists: {
        exists: true,
      },
    });
  });
};

function checkNetwork(options, cb) {
  var tasks = [
    checkRegisterHost.bind(null, options),
    checkMemberSSH.bind(null, options),
  ];
  async.parallel(async.reflectAll(tasks), checkNetworkResults);
  function checkNetworkResults(err, results) {
    results = {
      registerHost: results[0],
      memberReachable: results[1],
    };
    cb(err, results);
  }
};

function checkRegisterHost(options, cb) {
  options.path = '/sys.wlpn.internal/collective/local/hosts/';
  options.controllerHost = options.controllerHost || os.hostname() ||
    'localhost';
  options.controllerPort = options.controllerPort || 9443;

  if (!options.libertyExists) {
    return cb({
      msg: chalk.magenta('WARNING: Liberty does not exist, skipping ' +
        'registerHost check'),
    });
  }

  if (!options.adminUser || !options.adminPass) {
    var message = chalk.magenta([
      'WARNING: variables not specified: ',
      !options.adminUser ? 'adminUser, ' : '',
      !options.adminPass ? 'adminPass, ' : '',
      'skipping registerHost check',
    ].join(''));

    return cb({
      msg: message,
    });
  }

  try {
    getData(options, function(err, data) {
      var results = {};
      if (err) {
        results.msg = chalk.yellow('WARNING: could not parse registerHost ' +
          'vars in controller repo');
        results.err = err;
        return cb(results);
      }

      data = _.chain(data)
      .pickBy(function(val, key){
        if (val) {
          return key;
        }
        return false;
      })
      .mapKeys(function(val, key) {
        if (val) {
          var regex = new RegExp('\/sys.wlpn.internal\/collective\/local' +
            '\/hosts\/', 'gi');
          return key.replace(regex, '');
        }
      })
      .value();

      var message = [];

      if (/\b\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\b/gi
        .test(options.memberHost)) {
        message.push(chalk.yellow('WARNING: IP address specified for ' +
          'memberHost instead of a full-qualified hostname, did ' +
          'you use a fully-qualified hostname during registerHost?'));
      }

      if (_.isEmpty(data)) {
        message.push(chalk.red('ERROR: no environment variables detect, did ' +
          'you run registerHost on member host?'));
      }

      data = _.merge({},
        {
          msg: message.length > 0 ? message : 'Looks good',
        },
        {
          controllerHost: [options.controllerHost, options.controllerPort]
          .join(':'),
          memberHost: [options.memberHost, options.memberPort].join(':'),
        },
      data);

      // clean up option for next method to run
      delete options.path;

      return cb(null, data);
    });
  } catch (e) {
    return cb(e);
  }
}

function checkMemberSSH(options, cb) {
  options.memberHost = options.memberHost || os.hostname() || 'localhost';
  options.memberPort = options.memberPort || 22;
  if (!options.memberHost || !options.memberPort || !options.memberSSHUser ||
    !options.memberSSHPass) {
    var message = chalk.magenta([
      'WARNING: variables not specified: ',
      !options.memberHost ? 'memberHost, ' : '',
      !options.memberPort ? 'memberPort, ' : '',
      !options.memberSSHUser ? 'memberSSHUser, ' : '',
      !options.memberSSHPass ? 'memberSSHPass, ' : '',
      'skipping member reachability check',
    ].join(''));
    return cb({
      msg: message,
    });
  }

  var conn = new Client();
  conn.on('ready', function() {
    conn.end();
    return cb(null, {
      memberHost: [options.memberHost, ':', options.memberPort].join(''),
      success: true,
    });
  }).on('keyboard-interactive',
  function(name, instructions, instructionsLang, prompts, finish) {
    // OSX compatibility, doesn't support password auth, needs keyboard auth
    finish([options.memberSSHPass]);
  }).on('error', function(err) {
    return cb(null, {
      success: false,
      msg: chalk.red('ERROR: SSH failed, you will not be able to deploy to ' +
        'this member'),
      err: err,
    });
  }).connect({
    host: options.memberHost,
    port: options.memberPort,
    username: options.memberSSHUser,
    password: options.memberSSHPass,
    tryKeyboard: true,
  });
};

function getData(options, cb) {
  // This path can be a fixed constant in this module.
  var serverCommand = jmx.serverCommand;
  mandatory(options.controllerHost, 'controllerHost is missing');
  mandatory(options.controllerPort, 'controllerPort is missing');
  mandatory(options.adminUser, 'adminUser is missing');
  mandatory(options.adminPass, 'adminPass is missing');

  debug('Attempting request on controller host %s', options.host);
  var _jmx = options.jmx || jmx;
  var req = serverCommand.getSubTree(options);
  var endpoint = new _jmx.Endpoint(options.controllerHost,
    options.controllerPort, {
      acceptUnauthorized: true,
      username: options.adminPass,
      password: options.adminUser,
    });
  debug('The request being sent is: %s', JSON.stringify(req));
  endpoint.request(req, null, function(err, rsp, body) {
    if (err) {
      debug('The request returned with an error response: %s', err);
      // Create new error because we can't mutate default error, and Liberty's
      // call stack and body are going to be HUGE
      return cb(new Error(err.message));
    }
    return cb(null, rsp.body.value);
  });
};
